/*
 * balls.c
 *
 * GFreenect experiments
 *
 * Copyright (C) 2012, Eduardo Lima Mitev <elima@igalia.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * version 3, or (at your option) any later version as published by
 * the Free Software Foundation.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License at http://www.gnu.org/licenses/gpl-3.0.txt
 * for more details.
 */


/* PLEASE, NOTICE THAT THIS CODE IS STILL AN UNFINISHED, UNRELEASED EXPERIMENT */

#include <gfreenect.h>
#include <clutter/clutter.h>
#include <clutter-box2d.h>
#include <evd.h>
#include <cv.h>

#define WIDTH  640
#define HEIGHT 480

#define DEFAULT_N_LAYERS 2

static GFreenectDevice *kinect = NULL;
static ClutterActor *depth_tex;
static ClutterActor *video_tex;
static ClutterActor *stage;
static ClutterActor *stage1;
static ClutterActor *b2d_world;

static guchar *video_frame = NULL;
static guchar *video_frame1 = NULL;
static guchar *depth_frame_grayscale;
static guint16 *depth_frame_raw;

static gint count = 0;

static gint near_threshold = 100;
static gint far_threshold = 2700;

#define MAX_CONTOUR_BLOCKS 1000
static ClutterActor *contour_blocks[MAX_CONTOUR_BLOCKS] = { 0 , };
static GQueue *falling_items;

#define STEP 16
#define SCALE 1.0

typedef struct
{
  ClutterActor *tex;
  guchar *frame;
  gint index;
} FrameLayer;

static void process (guchar *video_frame, guint16 *depth_frame);

static gint n_layers = 0;
static FrameLayer **layers = NULL;

static guchar *new_buf;
static gsize new_buf_len = 640 * 480 / (STEP * STEP);

static void
quit (gint signale)
{
  signal (SIGINT, 0);

  g_signal_handlers_disconnect_by_func (stage, "quit", NULL);
  g_signal_handlers_disconnect_by_func (stage1, "quit", NULL);

  clutter_main_quit ();
}

static gboolean
inside_threshold (guint16 *depth_frame,  gint x, gint y)
{
  gint k;

  if (x < 0 || y < 0 || x >= WIDTH || y >= HEIGHT)
    return FALSE;

  k = y * WIDTH + x;

  return depth_frame[k] >= near_threshold && depth_frame[k] <= far_threshold;
}

static gboolean
is_contour (guint16 *depth_frame,  gint x, gint y)
{
  return
    (! inside_threshold (depth_frame, x - STEP, y)) ||
    (! inside_threshold (depth_frame, x + STEP, y)) ||
    (! inside_threshold (depth_frame, x, y - STEP)) ||
    (! inside_threshold (depth_frame, x, y + STEP));
}

static void
create_contour_block (ClutterActor **block, gint x, gint y, gint width, gint height)
{
  if ((*block) == NULL)
    {
      ClutterColor color = {0, 0, 0, 0};

      (*block) = clutter_rectangle_new_with_color (&color);

      clutter_actor_set_size ((*block), width, height);

      clutter_group_add (CLUTTER_GROUP (b2d_world), (*block));

      clutter_container_child_set (CLUTTER_CONTAINER (b2d_world), (*block),
                                   "manipulatable", FALSE,
                                   "friction", 0.5,
                                   "restitution", 1.0,
                                   "mode", CLUTTER_BOX2D_STATIC, NULL);

      g_object_ref ((*block));
    }

  clutter_actor_set_position ((*block), x, y);
}

static void
on_depth_frame (GFreenectDevice *kinect, gpointer user_data)
{
  guint16 *buf;
  guint8 *buf1;
  gsize len;
  GFreenectFrameMode frame_mode;
  GError *error = NULL;
  gint i;

  if (video_frame == NULL)
    return;

  /*
  count ++;
  if (count % 2 != 0)
    return;
  */

  buf = (guint16 *) gfreenect_device_get_depth_frame_raw (kinect, &len, &frame_mode);

  memcpy (depth_frame_raw, buf, 640 * 480 * 2);

  buf1 = gfreenect_device_get_depth_frame_grayscale (kinect, &len, &frame_mode);

  /*
  if (video_frame != NULL)
    {
      process (video_frame, depth_frame_raw);
      clutter_stage_queue_redraw (CLUTTER_STAGE (stage));
    }
  */

  /*
  for (i=0; i<640 * 480; i++)
    {
      if (buf[i] > far_threshold || buf[i] < near_threshold)
        {
          video_frame[i * 3 + 0] = 0;
          video_frame[i * 3 + 1] = 0;
          video_frame[i * 3 + 2] = 0;
        }
    }
  */

  gint j, k = 0;
  gint cb = 0;

  //  for (i=frame_mode.width-1; i>=0; i-=STEP)
  for (i=0; i<frame_mode.width; i+=STEP)
    for (j=0; j<frame_mode.height; j+=STEP)
      {
        guint p;
        guchar c;

        p = j * frame_mode.width + i;

        c = (guchar) ((depth_frame_raw[p] * 255) / far_threshold);

        if (inside_threshold (depth_frame_raw, i, j))
          {
            gint r = 255, g = 255, b = 255;

            new_buf[k] = c;

            if (is_contour (buf, i, j))
              {
                if (cb < MAX_CONTOUR_BLOCKS)
                  {
                    create_contour_block (&contour_blocks[cb], i, j, STEP * 2, STEP * 2);
                    cb ++;
                  }

                g = 0;
                b = 0;
              }

            depth_frame_grayscale[p * 3 + 0] = c;
            depth_frame_grayscale[p * 3 + 1] = c;
            depth_frame_grayscale[p * 3 + 2] = c;

            video_frame1[p * 3 + 0] = video_frame[p * 3 + 0];
            video_frame1[p * 3 + 1] = video_frame[p * 3 + 1];
            video_frame1[p * 3 + 2] = video_frame[p * 3 + 2];
          }
        else
          {
            new_buf[k] = 0;

            depth_frame_grayscale[p * 3 + 0] = 0;
            depth_frame_grayscale[p * 3 + 1] = 0;
            depth_frame_grayscale[p * 3 + 2] = 0;

            video_frame1[p * 3 + 0] = 0;
            video_frame1[p * 3 + 1] = 0;
            video_frame1[p * 3 + 2] = 0;
          }

        k++;
      }

  i = cb;
  while (contour_blocks[i] != NULL && i < MAX_CONTOUR_BLOCKS)
    {
      //      clutter_container_remove_actor (CLUTTER_CONTAINER (b2d_world), contour_blocks[i]);
      //      contour_blocks[i] = NULL;
      clutter_actor_set_position (contour_blocks[i], 10000, 10000);
      i++;
    }

  if (! clutter_texture_set_from_rgb_data (CLUTTER_TEXTURE (video_tex),
                                           video_frame,
                                           FALSE,
                                           640, 480,
                                           0,
                                           3,
                                           CLUTTER_TEXTURE_NONE,
                                           &error))
    {
      g_debug ("Error setting texture area: %s", error->message);
      g_error_free (error);
    }
  else
    {
      //      g_debug ("Video frame!");
    }

  if (! clutter_texture_set_from_rgb_data (CLUTTER_TEXTURE (depth_tex),
                                           depth_frame_grayscale,
                                           FALSE,
                                           640, 480,
                                           0,
                                           3,
                                           CLUTTER_TEXTURE_NONE,
                                           &error))
    {
      g_debug ("Error setting texture area: %s", error->message);
      g_error_free (error);
    }
  else
    {
      //      g_debug ("Depth frame!");
    }
}

static FrameLayer *
frame_layer_new (gint index)
{
  FrameLayer *layer;

  layer = g_slice_new0 (FrameLayer);
  layer->tex = clutter_texture_new ();

  clutter_actor_set_size (layer->tex, 640.0, 480.0);
  clutter_actor_set_position (layer->tex, 0.0, 0.0);
  clutter_actor_set_rotation (layer->tex,
                              CLUTTER_Y_AXIS,
                              180.0,
                              clutter_actor_get_width (layer->tex) / 2,
                              clutter_actor_get_height (layer->tex) / 2,
                              0);
  clutter_actor_set_depth (layer->tex, index*2);
  clutter_container_add_actor (CLUTTER_CONTAINER (stage), layer->tex);

  layer->frame = g_slice_alloc (640 * 480 * 4);

  layer->index = index;

  return layer;
}

static void
frame_layer_free (FrameLayer *layer)
{
  clutter_container_remove_actor (CLUTTER_CONTAINER (stage), layer->tex);
  g_object_unref (layer->tex);

  g_slice_free1 (WIDTH * HEIGHT * 4, layer->frame);

  g_slice_free (FrameLayer, layer);
}

static void
set_n_layers (gint n)
{
  gint i;

  if (layers != NULL)
    {
      for (i=0; i<n_layers; i++)
        frame_layer_free (layers[i]);

      g_slice_free1 (n_layers * sizeof (FrameLayer *), layers);
      layers = NULL;
    }

  n_layers = n;

  layers = g_slice_alloc (n_layers * sizeof (FrameLayer *));
  for (i=0; i<n_layers; i++)
    layers[i] = frame_layer_new (i);
}

static gint
get_frame_index_from_distance (guint16 d)
{
  gint inc;
  gint index;

  if (d < near_threshold)
    return n_layers - 1;

  inc = (far_threshold - near_threshold) / (n_layers - 1);

  index = (d - near_threshold) / inc;

  if (index >= n_layers)
    index = n_layers - 1;

  return index;
}

static void
process (guchar *video_frame, guint16 *depth_frame)
{
  gint i, j;
  GError *error = NULL;

  if (depth_frame == NULL)

  for (i=0; i<n_layers; i++)
    memset (layers[i]->frame, 0, 640 * 480 * 4);

  for (i=0; i<WIDTH * HEIGHT; i++)
    {
      guint16 d = depth_frame[i];
      gint index;
      FrameLayer *layer;

      index = get_frame_index_from_distance (d);
      //      index = n_layers - 1;
      layer = layers[index];

      layer->frame[i * 4 + 0] = video_frame[i * 3 + 0];
      layer->frame[i * 4 + 1] = video_frame[i * 3 + 1];
      layer->frame[i * 4 + 2] = video_frame[i * 3 + 2];
      layer->frame[i * 4 + 3] = 255;
    }

  for (i=0; i<n_layers; i++)
    if (! clutter_texture_set_from_rgb_data (CLUTTER_TEXTURE (layers[i]->tex),
                                             layers[i]->frame,
                                             TRUE,
                                             640, 480,
                                             0,
                                             4,
                                             CLUTTER_TEXTURE_NONE,
                                             &error))
      {
        g_debug ("Error setting texture area: %s", error->message);
        g_error_free (error);
        error = NULL;
      }
    else
      {
        //      g_debug ("Video frame!");
      }
}

static void
on_video_frame (GFreenectDevice *kinect, gpointer user_data)
{
  guchar *buf;
  GError *error = NULL;
  gsize len;
  GFreenectFrameMode frame_mode;
  gint i, k;

  buf = gfreenect_device_get_video_frame_raw (kinect, &len, &frame_mode);

  memcpy (video_frame, buf, len);

  /*
  if (! clutter_texture_set_from_rgb_data (CLUTTER_TEXTURE (video_tex),
                                           buf,
                                           FALSE,
                                           640, 480,
                                           0,
                                           3,
                                           CLUTTER_TEXTURE_NONE,
                                           &error))
    {
      g_debug ("Error setting texture area: %s", error->message);
      g_error_free (error);
    }
  else
    {
      //      g_debug ("Video frame!");
    }
  */
}

static void
on_new_kinect_device (GObject      *obj,
                      GAsyncResult *res,
                      gpointer      user_data)
{
  GError *error = NULL;

  kinect = gfreenect_device_new_finish (res, &error);
  if (kinect == NULL)
    {
      g_debug ("Failed to created kinect device: %s", error->message);
      g_error_free (error);

      quit (0);

      return;
    }

  g_debug ("Kinect device created!");

  g_signal_connect (kinect,
                    "depth-frame",
                    G_CALLBACK (on_depth_frame),
                    NULL);

  g_signal_connect (kinect,
                    "video-frame",
                    G_CALLBACK (on_video_frame),
                    NULL);

  gfreenect_device_set_tilt_angle (kinect, 0.0, NULL, NULL, NULL);

  gfreenect_device_start_depth_stream (kinect,
                                       GFREENECT_DEPTH_FORMAT_REGISTERED,
                                       &error);
  gfreenect_device_start_video_stream (kinect,
                                       GFREENECT_RESOLUTION_MEDIUM,
                                       GFREENECT_VIDEO_FORMAT_RGB,
                                       &error);
}

static void
paint_circle (ClutterActor *actor)
{
  /*
  gfloat radius = MIN (clutter_actor_get_width (actor),
                       clutter_actor_get_height (actor)) / 2.f;
  cogl_set_source_color4ub (0xe0, 0xe0, 0xe0, 0xff);
  cogl_path_arc (radius, radius, radius, radius, 0, 360);
  cogl_path_fill ();
  */

  if (clutter_actor_get_y (actor) >= 480.0 * 2.0)
    {
      clutter_container_remove_actor (CLUTTER_CONTAINER (b2d_world), actor);

      g_queue_push_tail (falling_items, actor);
    }
}

static ClutterActor *
create_falling_item ()
{
  ClutterActor *actor;

  actor = clutter_texture_new_from_file ("igalia.png", NULL);

  g_signal_connect (actor, "paint", G_CALLBACK (paint_circle), NULL);
  g_object_ref (actor);

  return actor;
}

static gboolean
create_falling_items (gpointer user_data)
{
  ClutterActor *item;
  gint radius;

  item = CLUTTER_ACTOR (g_queue_pop_head (falling_items));
  if (item == NULL)
    item = create_falling_item ();


  clutter_group_add (CLUTTER_CONTAINER (b2d_world), item);
  clutter_container_child_set (CLUTTER_CONTAINER (b2d_world), item,
                               "manipulatable", FALSE,
                               "friction", 0.8,
                               "is-circle", TRUE,
                               "mode", CLUTTER_BOX2D_DYNAMIC, NULL);

  radius = g_random_int_range (10, 40);
  clutter_actor_set_size (item, radius * 2, radius * 2);
  clutter_actor_set_position (item, radius + g_random_int_range (0, clutter_actor_get_width (video_tex) - radius * 2),
                              - radius * 2);

  g_timeout_add (g_random_int_range (20, 3000),
                 create_falling_items,
                 NULL);

  return FALSE;
}

gint
main (int argc, char *argv[])
{
  ClutterColor white = {255, 255, 255, 255};
  ClutterVertex gravity_vertex = {0.0, 10.0, 0.0};

  if (clutter_init (&argc, &argv) != CLUTTER_INIT_SUCCESS)
    return -1;

  gfreenect_device_new (0,
                        GFREENECT_SUBDEVICE_CAMERA,
                        NULL,
                        on_new_kinect_device,
                        NULL);

  new_buf = g_slice_alloc (new_buf_len);
  video_frame = g_slice_alloc (640 * 480 * 3);
  video_frame1 = g_slice_alloc (640 * 480 * 3);
  memset (video_frame1, 0, 640 * 480 * 3);
  depth_frame_raw = g_slice_alloc (640 * 480 * 2);
  depth_frame_grayscale = g_slice_alloc (640 * 480 * 3);

  falling_items = g_queue_new ();

  signal (SIGINT, quit);

  stage = clutter_stage_get_default ();
  g_signal_connect (stage, "destroy", G_CALLBACK (quit), NULL);
  clutter_actor_set_size (stage, 640 * SCALE * 2, 480 * SCALE);
  clutter_stage_set_title (CLUTTER_STAGE (stage), "GFreenect + Clutter + Box2D");
  //  clutter_stage_set_color (CLUTTER_STAGE (stage), &white);

  stage1 = clutter_stage_new ();
  g_signal_connect (stage1, "destroy", G_CALLBACK (quit), NULL);
  clutter_actor_set_size (stage1, 640, 480);

  //  set_n_layers (DEFAULT_N_LAYERS);

  b2d_world = clutter_box2d_new ();
  clutter_container_add_actor (CLUTTER_CONTAINER (stage), b2d_world);
  clutter_box2d_set_gravity (CLUTTER_BOX2D (b2d_world), &gravity_vertex);
  clutter_box2d_set_simulating (CLUTTER_BOX2D (b2d_world), TRUE);
  clutter_actor_set_rotation (b2d_world,
                              CLUTTER_Y_AXIS,
                              180.0,
                              (640 * SCALE) / 2,
                              (480 * SCALE) / 2,
                              0);
  //  clutter_actor_set_scale (b2d_world, 2.0, 2.0);

  depth_tex = clutter_texture_new ();
  g_object_ref (depth_tex);
  clutter_actor_set_size (depth_tex, 640.0, 480.0);
  clutter_actor_set_position (depth_tex, 640.0, 0.0);
  clutter_container_add_actor (CLUTTER_CONTAINER (stage), depth_tex);
  clutter_actor_set_rotation (depth_tex,
                              CLUTTER_Y_AXIS,
                              180.0,
                              clutter_actor_get_width (depth_tex) / 2,
                              clutter_actor_get_height (depth_tex) / 2,
                              0);

  video_tex = clutter_texture_new ();
  g_object_ref (video_tex);
  clutter_actor_set_size (video_tex, 640.0 * SCALE, 480.0 * SCALE);
  clutter_actor_set_position (video_tex, 0.0, 0.0);
  clutter_group_add (CLUTTER_CONTAINER (b2d_world), video_tex);
  /*
  clutter_actor_set_rotation (video_tex,
                              CLUTTER_Y_AXIS,
                              180.0,
                              clutter_actor_get_width (video_tex) / 2,
                              clutter_actor_get_height (video_tex) / 2,
                              0);
  */

  /*
  ClutterActor *title;
  ClutterColor white = {0xff, 0xff, 0xff, 0xff};

  title = clutter_text_new_full ("Sans 40px", "Clutter-Box2D", &white);
  clutter_actor_set_position (title, 100, 120);
  clutter_actor_set_rotation (title, CLUTTER_Z_AXIS, 23, 0, 0, 0);
  clutter_group_add (CLUTTER_GROUP (b2d_world), title);
  clutter_container_child_set (CLUTTER_CONTAINER (b2d_world), title,
                               "manipulatable", TRUE,
                               "mode", CLUTTER_BOX2D_DYNAMIC, NULL);
  */

  create_falling_items (NULL);

  clutter_actor_show_all (stage);
  //  clutter_actor_show_all (stage1);

  clutter_main ();

  if (kinect != NULL)
    g_object_unref (kinect);

  g_slice_free1 (new_buf_len, new_buf);
  g_slice_free1 (640 * 480 * 3, video_frame);
  g_slice_free1 (640 * 480 * 3, depth_frame_grayscale);
  g_slice_free1 (640 * 480 * 2, depth_frame_raw);

  return 0;
}
