/*
 * balls.c
 *
 * GFreenect experiments
 *
 * Copyright (C) 2012, Eduardo Lima Mitev <elima@igalia.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * version 3, or (at your option) any later version as published by
 * the Free Software Foundation.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License at http://www.gnu.org/licenses/gpl-3.0.txt
 * for more details.
 */


/* PLEASE, NOTICE THAT THIS CODE IS STILL AN UNFINISHED, UNRELEASED EXPERIMENT */

#include <gfreenect.h>
#include <clutter/clutter.h>
#include <evd.h>
#include <cv.h>

static GFreenectDevice *kinect = NULL;
static ClutterActor *depth_tex;
static ClutterActor *video_tex;

static EvdPeerManager *peer_manager;
static gint count = 0;
static EvdWebsocketServer *ws;

#define STEP 16

static guchar *new_buf;
static gsize new_buf_len = 640 * 480 / (STEP * STEP);

static void
quit (gint signale)
{
  signal (SIGINT, 0);

  clutter_main_quit ();
}

static void
on_depth_frame (GFreenectDevice *kinect, gpointer user_data)
{
  guint16 *buf;
  gsize len;
  GFreenectFrameMode frame_mode;
  GList *peers, *node;

  count ++;
  if (count % 2 != 0)
    return;

  buf = (guint16 *) gfreenect_device_get_depth_frame_raw (kinect, &len, &frame_mode);

  gint i, j;
  gint k = 0;
  for (i=frame_mode.width-1; i>=0; i-=STEP)
    for (j=0; j<frame_mode.height; j+=STEP)
      {
        guint p;
        guchar c;

        p = j * frame_mode.width + i;

        c = (guchar) ((buf[p] * 255) / 2047);

        if (buf[p] < 950)
          new_buf[k] = c;
        else
          new_buf[k] = 0;

        k++;
      }

  peers = evd_peer_manager_get_all_peers (peer_manager);
  node = peers;

  while (node != NULL)
    {
      EvdPeer *peer;

      peer = EVD_PEER (node->data);

      evd_peer_send (peer, (gchar *) new_buf, new_buf_len, NULL);

      node = node->next;
    }

  g_list_free (peers);

  /*
  if (! clutter_texture_set_from_rgb_data (CLUTTER_TEXTURE (depth_tex),
                                           buf,
                                           FALSE,
                                           640, 480,
                                           0,
                                           3,
                                           CLUTTER_TEXTURE_NONE,
                                           &error))
    {
      g_debug ("Error setting texture area: %s", error->message);
      g_error_free (error);
    }
  else
    {
      //      g_debug ("Depth frame!");
    }
  */
}

static void
on_video_frame (GFreenectDevice *kinect, gpointer user_data)
{
  guchar *buf;
  GError *error = NULL;
  gsize len;
  GFreenectFrameMode frame_mode;
  GList *peers, *node;

  //  count ++;
  if (count % 3 != 0)
    return;

  buf = gfreenect_device_get_video_frame_raw (kinect, &len, &frame_mode);

  peers = evd_peer_manager_get_all_peers (peer_manager);
  node = peers;

  while (node != NULL)
    {
      EvdPeer *peer;

      peer = EVD_PEER (node->data);

      evd_peer_send (peer, (gchar *) buf, len, NULL);

      node = node->next;
    }

  g_list_free (peers);

  /*
  if (! clutter_texture_set_from_rgb_data (CLUTTER_TEXTURE (video_tex),
                                           buf,
                                           FALSE,
                                           640, 480,
                                           0,
                                           3,
                                           CLUTTER_TEXTURE_NONE,
                                           &error))
    {
      g_debug ("Error setting texture area: %s", error->message);
      g_error_free (error);
    }
  else
    {
      //      g_debug ("Video frame!");
    }
  */
}

static void
on_new_kinect_device (GObject      *obj,
                      GAsyncResult *res,
                      gpointer      user_data)
{
  GError *error = NULL;

  kinect = gfreenect_device_new_finish (res, &error);
  if (kinect == NULL)
    {
      g_debug ("Failed to created kinect device: %s", error->message);
      g_error_free (error);

      //      quit (0);

      return;
    }

  g_debug ("Kinect device created!");

  g_signal_connect (kinect,
                    "depth-frame",
                    G_CALLBACK (on_depth_frame),
                    NULL);

  g_signal_connect (kinect,
                    "video-frame",
                    G_CALLBACK (on_video_frame),
                    NULL);

  gfreenect_device_set_tilt_angle (kinect, 5.0, NULL, NULL, NULL);

  gfreenect_device_start_depth_stream (kinect,
                                       GFREENECT_RESOLUTION_MEDIUM,
                                       &error);
  gfreenect_device_start_video_stream (kinect,
                                       GFREENECT_RESOLUTION_MEDIUM,
                                       GFREENECT_VIDEO_FORMAT_RGB,
                                       &error);
}

static void
on_new_peer (EvdTransport *transport, EvdPeer *peer, gpointer user_data)
{
  g_debug ("New peer: %s", evd_peer_get_id (peer));
}

int
main (int argc, char *argv[])
{
  ClutterActor *stage;
  EvdWebDir *web_dir;

  if (clutter_init (&argc, &argv) != CLUTTER_INIT_SUCCESS)
    return -1;

  ws = evd_websocket_server_new ();
  evd_websocket_server_set_standalone (ws, TRUE);
  evd_web_service_set_origin_policy (EVD_WEB_SERVICE (ws), EVD_POLICY_ALLOW);

  gfreenect_device_new (0,
                        GFREENECT_SUBDEVICE_CAMERA,
                        NULL,
                        on_new_kinect_device,
                        NULL);

  g_signal_connect (ws,
                    "new-peer",
                    G_CALLBACK (on_new_peer),
                    NULL);

  evd_service_listen (EVD_SERVICE (ws),
                      "0.0.0.0:8000",
                      NULL,
                      NULL,
                      NULL);

  peer_manager = evd_peer_manager_get_default ();

  web_dir = evd_web_dir_new ();
  evd_web_dir_set_root (web_dir, ".");

  evd_service_listen (EVD_SERVICE (web_dir),
                      "0.0.0.0:8080",
                      NULL,
                      NULL,
                      NULL);

  new_buf = g_slice_alloc (new_buf_len);

  signal (SIGINT, quit);

  stage = clutter_stage_get_default ();
  g_signal_connect (stage, "destroy", clutter_main_quit, NULL);
  clutter_actor_set_size (stage, 1280, 900);

  depth_tex = clutter_texture_new ();
  clutter_actor_set_size (depth_tex, 640.0, 480.0);
  clutter_container_add_actor (CLUTTER_CONTAINER (stage), depth_tex);

  video_tex = clutter_texture_new ();
  clutter_actor_set_size (video_tex, 640.0, 480.0);
  clutter_actor_set_position (video_tex, 650.0, 0.0);
  clutter_container_add_actor (CLUTTER_CONTAINER (stage), video_tex);

  //  clutter_actor_show_all (stage);

  clutter_main ();

  g_slice_free1 (new_buf_len, new_buf);

  g_object_unref (ws);
  g_object_unref (web_dir);
  g_object_unref (peer_manager);

  if (kinect != NULL)
    g_object_unref (kinect);

  return 0;
}
