/*
 * touch-surface.c
 *
 * GFreenect experiments
 *
 * Copyright (C) 2012, Eduardo Lima Mitev <elima@igalia.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * version 3, or (at your option) any later version as published by
 * the Free Software Foundation.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License at http://www.gnu.org/licenses/gpl-3.0.txt
 * for more details.
 */

/* PLEASE, NOTICE THAT THIS CODE IS STILL AN UNFINISHED, UNRELEASED EXPERIMENT */

#include <clutter/clutter.h>
#include <gfreenect.h>
#include <cv.h>
#include <string.h>
#include <math.h>
#include <evd.h>

#define WIDTH 640
#define HEIGHT 480

/*
static gint near_threshold = 950;
static gint far_threshold = 1001;
*/

static gint near_threshold = 2200;
static gint far_threshold = 2280;

static gint min_contour_area = 1.0;
static gint max_contour_area = 2000.0;

static gfloat point_proximity_threshold = 50.0;

static gint point_radius = 8.0;

static ClutterActor *stage;
static ClutterActor *group;
static GFreenectDevice *kinect = NULL;
static ClutterActor *depth_tex;
static ClutterActor *video_tex;

static guint8 *video_frame = NULL;
static guint8 *depth_frame_binary = NULL;
static guint8 *depth_frame_grayscale = NULL;
static guint16 *depth_frame = NULL;

static IplImage *cv_image = NULL;
static CvMemStorage *contours;
static CvMoments *moments;

static ClutterActor *roi_area = NULL;
static CvRect roi;

static GList *points = NULL;

static void
paint_circle (ClutterActor *actor)
{
  gfloat radius = MIN (clutter_actor_get_width (actor),
                       clutter_actor_get_height (actor)) / 2.f;

  if (radius < 10.0)
    cogl_set_source_color4ub (0xFF, 0x00, 0x00, 0xFF);
  else
    cogl_set_source_color4ub (0x00, 0xFF, 0x00, 200);
  cogl_path_arc (radius, radius, radius, radius, 0, 360);
  cogl_path_fill ();
  //  cogl_path_stroke ();
}

static ClutterActor *
create_point (gint x, gint y, gint radius)
{
  ClutterActor *circle;
  ClutterColor red = {0x00, 0x00, 0x00, 0x00};

  circle = clutter_rectangle_new_with_color (&red);
  clutter_actor_set_position (circle, x - radius, y - radius);
  clutter_actor_set_size (circle, radius * 2, radius * 2);
  g_signal_connect (circle, "paint",
                    G_CALLBACK (paint_circle), NULL);

  return circle;
}

static gboolean
check_point_threshold (gfloat x, gfloat y)
{
  GList *node;

  node = points;
  while (node != NULL)
    {
      ClutterActor *circle;
      gfloat cx, cy, cw, ch;
      gfloat dist;

      circle = CLUTTER_ACTOR (node->data);

      clutter_actor_get_position (circle, &cx, &cy);
      clutter_actor_get_size (circle, &cw, &ch);

      cx -= cw / 2;
      cy -= ch / 2;

      dist = sqrt (pow (abs (cx - x), 2) + pow (abs (cy - y), 2));
      //      g_debug ("%f", dist);

      if (dist < point_proximity_threshold)
        return FALSE;

      node = node->next;
    }

  return TRUE;
}

static void
get_closest_point_in_contour (CvRect bounding_rect, CvPoint *point)
{
  gint i, j;
  gint min = 100000;
  guint16 max = 0;

  point->x = 0;
  point->y = 0;

  for (i=0; i<bounding_rect.width; i++)
    for (j=0; j < bounding_rect.height; j++)
      {
        gint p;
        gint x, y;

        x = bounding_rect.x + i + roi.x;
        y = bounding_rect.y + j + roi.y;

        p = (y * WIDTH + x);

        if (depth_frame[p] >= near_threshold && depth_frame[p] <= far_threshold &&
            //            depth_frame[p] < min)
            depth_frame[p] > max)
          {
            //            min = depth_frame[p];
            max = depth_frame[p];

            point->x = x;
            point->y = y;
          }
      }
}

static void
notify_tap (gint x, gint y)
{
  gchar *msg;
  EvdPeerManager *peer_manager;
  GList *peers, *node;
  gfloat xp, yp;

  xp = (100 * (x - roi.x)) / roi.width;
  yp = (100 * (y - roi.y)) / roi.height;

  msg = g_strdup_printf ("[%f,%f]", xp, yp);

  peer_manager = evd_peer_manager_get_default ();

  peers = evd_peer_manager_get_all_peers (peer_manager);
  node = peers;

  while (node != NULL)
    {
      EvdPeer *peer;

      peer = EVD_PEER (node->data);

      evd_peer_send_text (peer, msg, NULL);

      node = node->next;
    }

  g_list_free (peers);
  g_free (msg);
  g_object_unref (peer_manager);
}

static void
find_contours (void *binary_frame)
{
  CvSeq *first_contour = NULL;
  gint nc;
  CvPoint offset = {0, 0};

  cvSetData (cv_image, binary_frame, WIDTH);
  cvSetImageROI (cv_image, roi);

  nc = cvFindContours (cv_image,
                       contours,
                       &first_contour,
                       sizeof (CvContour),
                       CV_RETR_EXTERNAL,
                       CV_CHAIN_APPROX_SIMPLE,
                       offset);

  if (nc <= 0)
    return;

  CvSeq *c;

  for (c=first_contour; c != NULL; c = c->h_next)
    {
      gfloat area;

      area = cvContourArea (c, CV_WHOLE_SEQ, FALSE);
      if (area >= min_contour_area && area <= max_contour_area)
        {
          gfloat center_x, center_y;
          gint cx, cy;
          gint p;
          ClutterActor *point;
          CvRect bounding_rect;
          CvPoint target;
          ClutterActor *rect;

          bounding_rect = cvBoundingRect (c, 1);

          /*
          g_debug ("%d, %d - %d, %d",
                   bounding_rect.x,
                   bounding_rect.y,
                   bounding_rect.width,
                   bounding_rect.height);
          */

          /*
          rect = clutter_rectangle_new ();
          clutter_actor_set_size (rect, bounding_rect.width, bounding_rect.height);
          clutter_actor_set_position (rect, bounding_rect.x + roi.x, bounding_rect.y + roi.y);
          clutter_container_add_actor (CLUTTER_CONTAINER (group), rect);
          points = g_list_append (points, rect);
          */

          get_closest_point_in_contour (bounding_rect, &target);

          /*
          cvMoments (c, moments, 1);

          center_x = (moments->m10 / moments->m00);
          center_y = (moments->m01 / moments->m00);
          */

          if (check_point_threshold (target.x, target.y))
            {

              /*
              cx = floor (center_x);
              cy = floor (center_y);
              */

              cx = target.x;
              cy = target.y;

              notify_tap (cx, cy);

              //              g_debug ("%d,%d", cx, cy);

              p = cy * WIDTH + cx;

              gint radius;

              //              radius = (depth_frame[p] - near_threshold) / 10;
              radius = 10;
              point = create_point (cx, cy, radius);
              clutter_container_add_actor (CLUTTER_CONTAINER (group), point);
              points = g_list_append (points, point);
            }
        }
    }

  //  g_debug ("# of contours: %d", nc);
}

static void
on_depth_frame (GFreenectDevice *kinect, gpointer user_data)
{
  guint max_points = 8;

  guint16 *buf;
  guint8 *buf1;
  gsize len;
  GFreenectFrameMode frame_mode;
  GError *error = NULL;
  gint i, j, p;

  if (video_frame == NULL)
    return;

  buf = (guint16 *) gfreenect_device_get_depth_frame_raw (kinect, &len, &frame_mode);
  buf1 = (guint16 *) gfreenect_device_get_depth_frame_grayscale (kinect, &len, NULL);

  depth_frame_grayscale = buf1;

  //  memset (depth_frame, 0, WIDTH * HEIGHT * 2);

  for (i=0; i<roi.width * roi.height; i++)
    {
      p = i + (roi.y * WIDTH) + (i / roi.width) * (WIDTH - roi.width) + roi.x;

      if (buf[p] >= near_threshold && buf[p] <= far_threshold)
        {
          depth_frame_binary[p] = 1;
          depth_frame[p] = buf[p];
        }
      else
        {
          /*
          buf1[p * 3 + 0] = 0;
          buf1[p * 3 + 1] = 0;
          buf1[p * 3 + 2] = 0;
          */

          video_frame[p * 3 + 0] -= 100;
          video_frame[p * 3 + 1] -= 100;
          video_frame[p * 3 + 2] -= 100;

          depth_frame_binary[p] = 0;
          depth_frame[p] = 0;
        }
    }

  g_list_free_full (points, (GDestroyNotify) clutter_actor_destroy);
  points = NULL;

  find_contours (depth_frame_binary);

  if (! clutter_texture_set_from_rgb_data (CLUTTER_TEXTURE (depth_tex),
                                           buf1,
                                           FALSE,
                                           640, 480,
                                           0,
                                           3,
                                           CLUTTER_TEXTURE_NONE,
                                           &error))
    {
      g_debug ("Error setting texture area: %s", error->message);
      g_error_free (error);
    }
  else
    {
      //      g_debug ("Depth frame!");
    }

  if (! clutter_texture_set_from_rgb_data (CLUTTER_TEXTURE (video_tex),
                                           video_frame,
                                           FALSE,
                                           640, 480,
                                           0,
                                           3,
                                           CLUTTER_TEXTURE_NONE,
                                           &error))
    {
      g_debug ("Error setting texture area: %s", error->message);
      g_error_free (error);
    }
  else
    {
      //      g_debug ("Video frame!");
    }
}

static void
on_video_frame (GFreenectDevice *kinect, gpointer user_data)
{
  guchar *buf;
  GError *error = NULL;
  gsize len;
  GFreenectFrameMode frame_mode;

  buf = gfreenect_device_get_video_frame_raw (kinect, &len, &frame_mode);

  memcpy (video_frame, buf, len);

  /*
  if (! clutter_texture_set_from_rgb_data (CLUTTER_TEXTURE (video_tex),
                                           buf,
                                           FALSE,
                                           640, 480,
                                           0,
                                           3,
                                           CLUTTER_TEXTURE_NONE,
                                           &error))
    {
      g_debug ("Error setting texture area: %s", error->message);
      g_error_free (error);
    }
  else
    {
      //      g_debug ("Video frame!");
    }
  */
}

static void
on_new_kinect_device (GObject      *obj,
                      GAsyncResult *res,
                      gpointer      user_data)
{
  GError *error = NULL;

  kinect = gfreenect_device_new_finish (res, &error);
  if (kinect == NULL)
    {
      g_debug ("Failed to created kinect device: %s", error->message);
      g_error_free (error);

      clutter_main_quit ();

      return;
    }

  g_debug ("Kinect device created!");

  g_signal_connect (kinect,
                    "depth-frame",
                    G_CALLBACK (on_depth_frame),
                    NULL);

  g_signal_connect (kinect,
                    "video-frame",
                    G_CALLBACK (on_video_frame),
                    NULL);

  gfreenect_device_set_tilt_angle (kinect, 0.0, NULL, NULL, NULL);

  gfreenect_device_start_depth_stream (kinect,
                                       GFREENECT_DEPTH_FORMAT_REGISTERED,
                                       &error);

  gfreenect_device_start_video_stream (kinect,
                                       GFREENECT_RESOLUTION_MEDIUM,
                                       GFREENECT_VIDEO_FORMAT_RGB,
                                       &error);
}

gint
main (gint argc, gchar *argv[])
{
  CvSize image_size;
  EvdWebDir *web_dir;
  EvdWebsocketServer *websocket;

  /* init */
  if (clutter_init (&argc, &argv) != CLUTTER_INIT_SUCCESS)
    return -1;

  video_frame = g_slice_alloc (WIDTH * HEIGHT * 3);
  depth_frame_binary = g_slice_alloc (WIDTH * HEIGHT);
  depth_frame = g_slice_alloc (WIDTH * HEIGHT * 2);

  image_size.width = WIDTH;
  image_size.height = HEIGHT;
  cv_image = cvCreateImage (image_size, IPL_DEPTH_8U, 1);
  contours = cvCreateMemStorage (0);
  moments = g_slice_new (CvMoments);

  points = g_queue_new ();

  roi.x = 90;
  roi.y = 80;
  roi.width = 420;
  roi.height = 290;

  cvSetImageROI (cv_image, roi);

  /* web dir */
  /*
  web_dir = evd_web_dir_new ();
  evd_web_dir_set_root (web_dir, "/home/elima/lab/processing-js");
  evd_service_listen (EVD_SERVICE (web_dir),
                      "0.0.0.0:8080",
                      NULL,
                      NULL,
                      NULL);
  */

  websocket = evd_websocket_server_new ();
  evd_web_service_set_origin_policy (EVD_WEB_SERVICE (websocket), EVD_POLICY_ALLOW);
  evd_websocket_server_set_standalone (websocket, TRUE);
  evd_service_listen (EVD_SERVICE (websocket),
                      "0.0.0.0:8081",
                      NULL,
                      NULL,
                      NULL);

  /* kinect device */
  gfreenect_device_new (0,
                        GFREENECT_SUBDEVICE_CAMERA,
                        NULL,
                        on_new_kinect_device,
                        NULL);

  /* stage */
  stage = clutter_stage_get_default ();
  g_signal_connect (stage, "destroy", clutter_main_quit, NULL);
  clutter_actor_set_size (stage, WIDTH * 2 + 10, HEIGHT);
  clutter_stage_set_title (CLUTTER_STAGE (stage), "Touch Surface calibration tool");

  group = clutter_group_new ();
  clutter_container_add_actor (CLUTTER_CONTAINER (stage), group);
  /*
  clutter_actor_set_rotation (group,
                              CLUTTER_Y_AXIS,
                              180.0,
                              clutter_actor_get_width (stage) / 2,
                              clutter_actor_get_height (stage) / 2,
                              0);
  */

  /* depth texture */
  depth_tex = clutter_texture_new ();
  clutter_actor_set_size (depth_tex, WIDTH, HEIGHT);
  clutter_actor_set_position (depth_tex, WIDTH + 10, 0.0);
  clutter_container_add_actor (CLUTTER_CONTAINER (group), depth_tex);
  /*
  clutter_actor_set_rotation (depth_tex,
                              CLUTTER_Y_AXIS,
                              180.0,
                              clutter_actor_get_width (depth_tex) / 2,
                              clutter_actor_get_height (depth_tex) / 2,
                              0);
  */

  /* video texture */
  video_tex = clutter_texture_new ();
  clutter_actor_set_size (video_tex, 640.0, 480.0);
  clutter_actor_set_position (video_tex, 0.0, 0.0);
  clutter_container_add_actor (CLUTTER_CONTAINER (group), video_tex);
  /*
  clutter_actor_set_rotation (video_tex,
                              CLUTTER_Y_AXIS,
                              180.0,
                              clutter_actor_get_width (video_tex) / 2,
                              clutter_actor_get_height (video_tex) / 2,
                              0);
  */

  /* start the show */
  clutter_actor_show_all (stage);
  clutter_main ();

  /* free memory */
  if (kinect != NULL)
    g_object_unref (kinect);

  g_slice_free1 (WIDTH * HEIGHT * 3, video_frame);
  g_slice_free1 (WIDTH * HEIGHT, depth_frame_binary);
  g_slice_free1 (WIDTH * HEIGHT * 2, depth_frame);
  g_slice_free (CvMoments, moments);

  g_object_unref (websocket);

  /* @TODO: check why this segfaults */
  //  cvReleaseImage (&cv_image);

  cvReleaseMemStorage (&contours);

  return 0;
}
